/*
 * Copyright (C) 2007 NetXen Inc.
 * 
 * All rights reserved.
 * 
 * The code contained in this file has been developed by NetXen Inc.
 * 
 * By downloading, compiling or using the code in any form, you agree to be bound
 * by the following terms:
 * 
 * 1) You may compile the unmodified source code file and reproduce and install
 * the resulting binary code file solely for your internal use with NetXen Inc.
 * hardware and systems you have purchased.
 * 
 * 2) Except as expressly permitted herein or in a separately signed agreement,
 * any other use, reproduction, distribution, modification, display, creation
 * of derivative works or circulation of the code is strictly prohibited.  You
 * may not transfer the code or assign your rights or obligations hereunder
 * without the prior written consent of NetXen Inc.
 * 
 * 3) This file contains confidential, proprietary and trade secret information
 * of NetXen Inc. and may not be disclosed to any third party or used for any
 * other purpose, except as expressly permitted herein.
 * 
 * 4) THIS SOFTWARE IS PROVIDED “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETXEN INC.
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.  THE APPLICATION OF THE UNITED NATIONS CONVENTION
 * ON CONTRACTS FOR THE INTERNATIONAL SALE OF GOODS IS HEREBY DISCLAIMED.
 * 
 * If you do not agree to the foregoing terms, you may not download, use or
 * compile this code in any form and must promptly delete or destroy all copiesi
 * of the code you have downloaded, compiled or reproduced.
 * 
 * Contact Information:
 * licensing@netxen.com
 * NetXen Inc.
 * 3965 Freedom Circle, Fourth Floor
 * Santa Clara, CA 95054
 * 
 */

/* 
 * This is the intercept module that will arbitrate between offload and 
 * native stack. Since offload decision is taken in listen() and connect(),
 * they are unconditionally passed on to offload stack. For other calls, a 
 * check for offload is made.
 */
#include <linux/types.h>
#include <linux/module.h>
#include <linux/version.h>
#include <asm/atomic.h>
#include <linux/net.h>
#include <net/sock.h>
#include <linux/poll.h>
#include <linux/jhash.h>
#include <linux/inetdevice.h>
#include <net/ip.h>
#include <net/tcp.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,14)
#include <net/inet_hashtables.h>
#endif
#include "nx_intercept.h"
#include "kernel_compatibility.h"
#include "unm_version.h"
#include "unm_inc.h"
#include "netxen_linux_ioctl.h"

MODULE_AUTHOR("Copyright (C) 2003 - 2007 NetXen, Inc.");
MODULE_DESCRIPTION("NetXen Intercept Module");
MODULE_LICENSE("GPL");
MODULE_VERSION(UNM_NIC_LINUX_VERSIONID);	/* We build all 3 modules
						   together so the same
						   version is fine. */

extern struct proto_ops inet_stream_ops;
extern struct proto_ops inet_dgram_ops;

typedef struct {
	atomic_t		ref_count;
	struct proto_ops	lsa_stream_ops;
	struct proto_ops	host_stream_ops;
	nx_ioctl_t		lsa_dgram_ioctl;
	nx_ioctl_t		host_dgram_ioctl;
	int			is_registered;
} nx_lsa_intercept_t;

static nx_lsa_intercept_t	nx_lsa_inet;

/*
 *
 */
int nx_intercept_inet_release(struct socket *sock)
{
	if (UNM_SOCK_IS_SOCK_OFFLOADED(sock)) {
		return (nx_lsa_inet.lsa_stream_ops.release(sock));
	}

	return (nx_lsa_inet.host_stream_ops.release(sock));
}

/*
 *
 */
int nx_intercept_inet_socketpair(struct socket *sock1, struct socket *sock2)
{
	if (UNM_SOCK_IS_SOCK_OFFLOADED(sock1)) {
		return (nx_lsa_inet.lsa_stream_ops.socketpair(sock1, sock2));
	}

	return (nx_lsa_inet.host_stream_ops.socketpair(sock1, sock2));
}

/*
 *
 */
int nx_intercept_inet_accept(struct socket *sock, struct socket *newsock,
			     int flags)
{
	int rv;
	
	if (UNM_SOCK_IS_SOCK_OFFLOADED(sock)) {
		rv = nx_lsa_inet.lsa_stream_ops.accept(sock, newsock, flags);
	} else {
		rv =  nx_lsa_inet.host_stream_ops.accept(sock, newsock, flags);
		if (!rv && newsock->sk) { 
			newsock->sk->SK_PROTINFO = NULL;
		}
	}

	return rv;
	
}

/*
 *
 */
int nx_intercept_inet_getname(struct socket *sock, struct sockaddr *uaddr,
			      int *uaddr_len, int peer)
{
	if (UNM_SOCK_IS_SOCK_OFFLOADED(sock)) {
		return (nx_lsa_inet.lsa_stream_ops.getname(sock, uaddr,
							   uaddr_len, peer));
	}

	return (nx_lsa_inet.host_stream_ops.getname(sock, uaddr, uaddr_len,
						    peer));
}

/*
 *
 */
unsigned int nx_intercept_inet_poll(struct file *file, struct socket *sock,
				    poll_table *wait)
{
	if (UNM_SOCK_IS_SOCK_OFFLOADED(sock)) {
		return (nx_lsa_inet.lsa_stream_ops.poll(file, sock, wait));
	}

	return (nx_lsa_inet.host_stream_ops.poll(file, sock, wait));
}

/*
 *
 */
int nx_intercept_inet_ioctl(struct socket *sock, unsigned int cmd,
			    unsigned long arg)
{
	int rv;

	/* If it is SIOC_NETXEN_OFFLOAD, it is allowed even if the application
	   (e.g. nxoffload) is not offloaded. */
	if (cmd == SIOC_NETXEN_OFFLOAD) {
		if (nx_lsa_inet.is_registered) {
			atomic_inc(&nx_lsa_inet.ref_count);
			if (nx_lsa_inet.is_registered) {
				rv = nx_lsa_inet.
					lsa_stream_ops.ioctl(sock,cmd,arg);
				atomic_dec(&nx_lsa_inet.ref_count);
				return rv;
			}

			atomic_dec(&nx_lsa_inet.ref_count);
		} /* if is_registered */

		goto native;
	}

	if (UNM_SOCK_IS_SOCK_OFFLOADED(sock)) {
		/* no need for intercept's own lsa refcount here, the
		   offloaded sock should have taken lsa's module reference.
		 */
		return (nx_lsa_inet.lsa_stream_ops.ioctl(sock, cmd, arg));
	}

native:
	return (nx_lsa_inet.host_stream_ops.ioctl(sock, cmd, arg));
}

/*
 *
 */
int nx_intercept_inet_shutdown(struct socket *sock, int how)
{
	if (UNM_SOCK_IS_SOCK_OFFLOADED(sock)) {
		return (nx_lsa_inet.lsa_stream_ops.shutdown(sock, how));
	}

	return (nx_lsa_inet.host_stream_ops.shutdown(sock, how));
}

/*
 *
 */
int nx_intercept_inet_setsockopt(struct socket *sock, int level, int optname,
				 char *optval, int optlen)
{
	if (UNM_SOCK_IS_SOCK_OFFLOADED(sock)) {
		return (nx_lsa_inet.lsa_stream_ops.setsockopt(sock, level,
							      optname, optval,
							      optlen));
	}

	return (nx_lsa_inet.host_stream_ops.setsockopt(sock, level, optname,
						       optval, optlen));
}

/*
 *
 */
int nx_intercept_inet_getsockopt(struct socket *sock, int level, int optname,
				 char *optval, int *optlen)
{
	if (UNM_SOCK_IS_SOCK_OFFLOADED(sock)) {
		return (nx_lsa_inet.lsa_stream_ops.getsockopt(sock, level,
							      optname, optval,
							      optlen));
	}
	return (nx_lsa_inet.host_stream_ops.getsockopt(sock, level, optname,
						       optval, optlen));
}

/*
 *
 */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
static int nx_intercept_inet_sendmsg(struct kiocb *iocb, struct socket *sock,
				     struct msghdr *msg, size_t size)
#else
static int nx_intercept_inet_sendmsg(struct socket *sock, struct msghdr *msg,
				     size_t size, struct scm_cookie *scm)
#endif
{
	if (UNM_SOCK_IS_SOCK_OFFLOADED(sock)) {
		return (nx_lsa_inet.lsa_stream_ops.sendmsg(iocb, sock, msg,
							   size));
	}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	return (nx_lsa_inet.host_stream_ops.sendmsg(iocb, sock, msg, size));
#else
	return (nx_lsa_inet.host_stream_ops.sendmsg(sock, msg, size, scm));
#endif
}

/*
 *
 */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
static int nx_intercept_inet_recvmsg(struct kiocb *iocb, struct socket *sock,
				     struct msghdr *msg, size_t size,
				     int flags)
#else
static int nx_intercept_inet_recvmsg(struct socket *sock, struct msghdr *msg,
				     size_t size, int flags,
				     struct scm_cookie *scm)
#endif
{
	if (UNM_SOCK_IS_SOCK_OFFLOADED(sock)) {
		return (nx_lsa_inet.lsa_stream_ops.recvmsg(iocb, sock, msg,
							   size, flags));
	}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	return (nx_lsa_inet.host_stream_ops.recvmsg(iocb, sock, msg, size,
						    flags));
#else
	return (nx_lsa_inet.host_stream_ops.recvmsg(sock, msg, size, flags,
						    scm));
#endif
}

/*
 *
 */
ssize_t nx_intercept_inet_sendpage(struct socket *sock, struct page *page,
				   int offset, size_t size, int flags)
{
	if (UNM_SOCK_IS_SOCK_OFFLOADED(sock)) {
		return (nx_lsa_inet.lsa_stream_ops.sendpage(sock, page, offset,
							    size, flags));
	}

	return (nx_lsa_inet.host_stream_ops.sendpage(sock, page, offset, size,
						     flags));
}
#if (defined (CONFIG_COMPAT)) && (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,17))
/*
 *
 */
int nx_compat_sock_common_setsockopt(struct socket *sock, int level, int optname,
                                  char *optval, int optlen)
{
	if (UNM_SOCK_IS_SOCK_OFFLOADED(sock)) {
		return (nx_lsa_inet.lsa_stream_ops.compat_setsockopt(sock, level, optname,
							    optval, optlen));
	}

	return (nx_lsa_inet.host_stream_ops.compat_setsockopt(sock, level, optname,
    								optval, optlen));
}
/*
 *
 */
int nx_compat_sock_common_getsockopt(struct socket *sock, int level, int optname,
                                  char *optval, int *optlen)
{
	if (UNM_SOCK_IS_SOCK_OFFLOADED(sock)) {
		return (nx_lsa_inet.lsa_stream_ops.compat_getsockopt(sock, level, optname,
							    optval, optlen));
	}

	return (nx_lsa_inet.host_stream_ops.compat_getsockopt(sock, level, optname,
    								optval, optlen));
}
#endif
/*
 *
 */
int nx_intercept_inet_listen(struct socket *sock, int backlog)
{
	int rv;

	if (nx_lsa_inet.is_registered) {
		atomic_inc(&nx_lsa_inet.ref_count);
		if (nx_lsa_inet.is_registered) {
			rv = nx_lsa_inet.lsa_stream_ops.listen(sock, backlog);
		} else {
			atomic_dec(&nx_lsa_inet.ref_count);
			goto no_tnic;
		}
		atomic_dec(&nx_lsa_inet.ref_count);
	} else {
	  no_tnic:
		rv =  nx_lsa_inet.host_stream_ops.listen(sock, backlog);
	}

	return rv;
}

/*
 *
 */
int nx_intercept_inet_connect(struct socket *sock, struct sockaddr *uaddr,
				 int addr_len, int flags)
{
	int rv;

	if (nx_lsa_inet.is_registered) {
		atomic_inc(&nx_lsa_inet.ref_count);
		if (nx_lsa_inet.is_registered) {
			 rv = nx_lsa_inet.lsa_stream_ops.connect(sock, uaddr,
								 addr_len,
								 flags);
		} else {
			atomic_dec(&nx_lsa_inet.ref_count);
			goto no_tnic;
		}
		atomic_dec(&nx_lsa_inet.ref_count);
	} else {
	  no_tnic:
		 rv =  nx_lsa_inet.host_stream_ops.connect(sock, uaddr,
							   addr_len, flags);
	}

	return rv;

}

int nx_intercept_inet_dgram_ioctl(struct socket *sock, unsigned int cmd,
		unsigned long arg)
{
	int rv;

	if (cmd == SIOC_NETXEN_OFFLOAD) {
		if (nx_lsa_inet.is_registered) {
			atomic_inc(&nx_lsa_inet.ref_count);
			if (nx_lsa_inet.is_registered) {
				rv = nx_lsa_inet.lsa_dgram_ioctl(sock,cmd,arg);
				atomic_dec(&nx_lsa_inet.ref_count);
				return rv;
			}

			atomic_dec(&nx_lsa_inet.ref_count);
		} /* if is_registered */
	}

	return (nx_lsa_inet.host_dgram_ioctl(sock, cmd, arg));
}

/*
 *
 */
int nx_lsa_register_ops(nx_intercept_registry_t *registry,
			nx_intercept_info_t *ret_info)
{
	if (nx_lsa_inet.is_registered) {
		return -EINVAL;
	}

	nx_lsa_inet.lsa_stream_ops.release = registry->lsa_stream_ops.release;
	nx_lsa_inet.lsa_stream_ops.socketpair =
		registry->lsa_stream_ops.socketpair;
	nx_lsa_inet.lsa_stream_ops.accept = registry->lsa_stream_ops.accept;
	nx_lsa_inet.lsa_stream_ops.getname = registry->lsa_stream_ops.getname;
	nx_lsa_inet.lsa_stream_ops.poll = registry->lsa_stream_ops.poll;
	nx_lsa_inet.lsa_stream_ops.ioctl = registry->lsa_stream_ops.ioctl;
	nx_lsa_inet.lsa_stream_ops.shutdown =
		registry->lsa_stream_ops.shutdown;
	nx_lsa_inet.lsa_stream_ops.setsockopt =
		registry->lsa_stream_ops.setsockopt;
	nx_lsa_inet.lsa_stream_ops.getsockopt =
		registry->lsa_stream_ops.getsockopt;
	nx_lsa_inet.lsa_stream_ops.sendmsg = registry->lsa_stream_ops.sendmsg;
	nx_lsa_inet.lsa_stream_ops.recvmsg = registry->lsa_stream_ops.recvmsg;
	nx_lsa_inet.lsa_stream_ops.sendpage =
		registry->lsa_stream_ops.sendpage;
	nx_lsa_inet.lsa_stream_ops.listen = registry->lsa_stream_ops.listen;
	nx_lsa_inet.lsa_stream_ops.connect = registry->lsa_stream_ops.connect;
#if (defined (CONFIG_COMPAT)) && (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,17))	
	nx_lsa_inet.lsa_stream_ops.compat_setsockopt = registry->lsa_stream_ops.compat_setsockopt;
	nx_lsa_inet.lsa_stream_ops.compat_getsockopt = registry->lsa_stream_ops.compat_getsockopt;
#endif	

	nx_lsa_inet.lsa_dgram_ioctl = registry->lsa_dgram_ioctl;

	memcpy(&ret_info->host_stream_ops, &nx_lsa_inet.host_stream_ops,
	       sizeof (ret_info->host_stream_ops));
	ret_info->host_dgram_ioctl = nx_lsa_inet.host_dgram_ioctl;

	nx_lsa_inet.is_registered = 1;

	return 0;
}

/*
 *
 */
void nx_lsa_unregister_ops(void)
{
	nx_lsa_inet.is_registered = 0;

retry:
	/* wait for refcount to become 0 */
	if (atomic_read(&nx_lsa_inet.ref_count)) {

		printk("ref count not 0 :%d\n",
		       atomic_read(&nx_lsa_inet.ref_count));
		current->state = TASK_UNINTERRUPTIBLE;
		schedule_timeout(3);
		current->state = TASK_RUNNING;
		goto retry;
	}

	/* all done */
	memset(&nx_lsa_inet.lsa_stream_ops, 0,
		sizeof(nx_lsa_inet.lsa_stream_ops));
	nx_lsa_inet.lsa_dgram_ioctl = NULL;
	return;
}

/*
 *
 */
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16)  || \
    (LINUX_VERSION_CODE == KERNEL_VERSION(2,6,15) && (FEDORA==FC5))) && \
    defined(CONFIG_DEBUG_RODATA)
static int change_perm(unsigned long ptr, unsigned int size, pgprot_t prot)
{
#if (defined(RHEL5) && defined(HOST_CPU_x86_64))
#define delta(x, y) (((x) > (y)) ? (x) - (y) : (y) - (x))
#define pasym(ptr) ((unsigned long)(init_mm.pgd) - PAGE_OFFSET + delta((unsigned long)&init_level4_pgt, ptr))
	unsigned long va, phys, pfn_pa;
	unsigned long endptr = (ptr + size + PAGE_SIZE - 1) & PAGE_MASK;

	printk("%s: rhel5 workaround\n", __FUNCTION__);
	for (ptr &= PAGE_MASK; ptr < endptr; ptr += PAGE_SIZE) {
		phys = pasym(ptr);
		pfn_pa = phys >> PAGE_SHIFT;
		va = (unsigned long)__va(phys);
		if (change_page_attr(virt_to_page(va), 1, PAGE_KERNEL) < 0) {
			printk("Failed\n");
			return -1;
		}
	}
#else
	unsigned long endptr = (ptr + size + PAGE_SIZE - 1) & PAGE_MASK;
	size = (size + PAGE_SIZE - 1) >> PAGE_SHIFT;

	printk("%s: regular\n", __FUNCTION__);
	for (ptr &= PAGE_MASK; ptr < endptr; ptr += PAGE_SIZE) {
		if (change_page_attr(virt_to_page(ptr), 1, prot) < 0) {
			return -1;
		}
	}
#endif
	return 0;
}
#endif

/*
 *
 */
static int __init nx_intercept_init(void)
{
#if 1
	if (!try_module_get(THIS_MODULE))
		return -EINVAL;
#endif
	nx_lsa_inet.is_registered	= 0;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16)  || \
    (LINUX_VERSION_CODE == KERNEL_VERSION(2,6,15) && (FEDORA==FC5))) && \
    defined(CONFIG_DEBUG_RODATA)
        if (change_perm((unsigned long)&inet_stream_ops,
                         sizeof(inet_stream_ops),
                         PAGE_KERNEL_FLAG) < 0) {
		module_put(THIS_MODULE);
                return -EIO;
        }

        if (change_perm((unsigned long)&inet_dgram_ops,
                         sizeof(inet_dgram_ops),
                         PAGE_KERNEL_FLAG) < 0) {
		change_perm((unsigned long)&inet_stream_ops,
				sizeof(inet_stream_ops),
				PAGE_KERNEL_RO);
		module_put(THIS_MODULE);
                return -EIO;
        }
#endif

	memcpy(&nx_lsa_inet.host_stream_ops, &inet_stream_ops,
		sizeof(inet_stream_ops));
	nx_lsa_inet.host_dgram_ioctl = inet_dgram_ops.ioctl;

        inet_stream_ops.release = nx_intercept_inet_release;
        inet_stream_ops.connect = nx_intercept_inet_connect;
        inet_stream_ops.socketpair = nx_intercept_inet_socketpair;
        inet_stream_ops.accept = nx_intercept_inet_accept;
        inet_stream_ops.getname = nx_intercept_inet_getname;
        inet_stream_ops.poll = nx_intercept_inet_poll;
        inet_stream_ops.ioctl = nx_intercept_inet_ioctl;
        inet_stream_ops.listen = nx_intercept_inet_listen;
        inet_stream_ops.shutdown = nx_intercept_inet_shutdown;
        inet_stream_ops.setsockopt = nx_intercept_inet_setsockopt;
        inet_stream_ops.getsockopt = nx_intercept_inet_getsockopt;
        inet_stream_ops.sendmsg = nx_intercept_inet_sendmsg;
        inet_stream_ops.recvmsg = nx_intercept_inet_recvmsg;
        inet_stream_ops.sendpage = nx_intercept_inet_sendpage;
#if (defined (CONFIG_COMPAT)) && (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,17))	
        inet_stream_ops.compat_setsockopt = nx_compat_sock_common_setsockopt;
        inet_stream_ops.compat_getsockopt = nx_compat_sock_common_getsockopt;
#endif

	inet_dgram_ops.ioctl = nx_intercept_inet_dgram_ioctl;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16)  || \
    (LINUX_VERSION_CODE == KERNEL_VERSION(2,6,15) && (FEDORA==FC5))) && \
    defined(CONFIG_DEBUG_RODATA)
        change_perm((unsigned long)&inet_stream_ops, sizeof(inet_stream_ops),
                    PAGE_KERNEL_RO);
        change_perm((unsigned long)&inet_dgram_ops, sizeof(inet_dgram_ops),
                    PAGE_KERNEL_RO);
#endif

	atomic_set(&nx_lsa_inet.ref_count, 0);
	memset(&nx_lsa_inet.lsa_stream_ops, 0,
	       sizeof(nx_lsa_inet.lsa_stream_ops));
	nx_lsa_inet.lsa_dgram_ioctl = NULL;

	return 0;
}

/*
 *
 */
static void __exit nx_intercept_exit(void)
{
	/* this isn't allowed */
	BUG();

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16)  || \
    (LINUX_VERSION_CODE == KERNEL_VERSION(2,6,15) && (FEDORA==FC5))) && \
    defined(CONFIG_DEBUG_RODATA)
        if (change_perm((unsigned long)&inet_stream_ops,
                         sizeof(inet_stream_ops),
                         PAGE_KERNEL_FLAG) < 0) {
		return;
        }

        if (change_perm((unsigned long)&inet_dgram_ops,
                         sizeof(inet_dgram_ops),
                         PAGE_KERNEL_FLAG) < 0) {
		change_perm((unsigned long)&inet_stream_ops,
				sizeof(inet_stream_ops),
				PAGE_KERNEL_RO);
		return;
        }
#endif

	memcpy(&inet_stream_ops, &nx_lsa_inet.host_stream_ops,
		sizeof(inet_stream_ops));

	inet_dgram_ops.ioctl = nx_lsa_inet.host_dgram_ioctl;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16)  || \
    (LINUX_VERSION_CODE == KERNEL_VERSION(2,6,15) && (FEDORA==FC5))) && \
    defined(CONFIG_DEBUG_RODATA)
        change_perm((unsigned long)&inet_stream_ops, sizeof(inet_stream_ops),
                    PAGE_KERNEL_RO);
        change_perm((unsigned long)&inet_dgram_ops, sizeof(inet_dgram_ops),
                    PAGE_KERNEL_RO);
#endif

}

module_init(nx_intercept_init);
module_exit(nx_intercept_exit);

EXPORT_SYMBOL(nx_lsa_unregister_ops);
EXPORT_SYMBOL(nx_lsa_register_ops);


/****************************************************************************
 *
 *			To access the TCP stack information
 *
 *****************************************************************************/

#if	(LINUX_VERSION_CODE > KERNEL_VERSION(2,6,9)) &&	\
	(LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,13))
extern struct sock *tcp_v4_lookup(u32 saddr, u16 sport, u32 daddr, u16 dport,
				  int dif);
#endif

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,9)
/*
 *
 */
static __inline__ int nx_tcp_hashfn(__u32 laddr, __u16 lport,
				    __u32 faddr, __u16 fport)
{
	int h = (laddr ^ lport) ^ (faddr ^ fport);
	h ^= h >> 16;
	h ^= h >> 8;
	return h & (tcp_ehash_size - 1);
}

/*
 *
 */
static inline struct sock *nx_tcp_v4_lookup_established(u32 saddr, u16 sport,
                                                        u32 daddr, u16 hnum,
                                                        int dif)
{
	struct tcp_ehash_bucket *head;
	TCP_V4_ADDR_COOKIE(acookie, saddr, daddr)
		__u32 ports = TCP_COMBINED_PORTS(sport, hnum);
	struct sock *sk;
	struct hlist_node *node;
	/* Optimize here for direct hit, only listening connections can
	 * have wildcards anyways.
	 */
	int hash = nx_tcp_hashfn(daddr, hnum, saddr, sport);
	head = &tcp_ehash[hash];
	read_lock(&head->lock);
	sk_for_each(sk, node, &head->chain) {
		if (TCP_IPV4_MATCH(sk, acookie, saddr, daddr, ports, dif))
			goto hit; /* You sunk my battleship! */
	}

	/* Must check for a TIME_WAIT'er before going to listener hash. */
	sk_for_each(sk, node, &(head + tcp_ehash_size)->chain) {
		if (TCP_IPV4_TW_MATCH(sk, acookie, saddr, daddr, ports, dif))
			goto hit;
	}
	sk = NULL;
  out:
	read_unlock(&head->lock);
	return sk;
  hit:
	sock_hold(sk);
	goto out;
}

/*
 *
 */
static inline struct sock *__nx_tcp_v4_lookup(u32 saddr, u16 sport,
                                            u32 daddr, u16 hnum, int dif)
{
	struct sock *sk = nx_tcp_v4_lookup_established(saddr, sport,
                                                       daddr, hnum, dif);

	return sk ? : tcp_v4_lookup_listener(daddr, hnum, dif);
}

/*
 *
 */
static inline struct sock *nx_tcp_v4_lookup(u32 saddr, u16 sport, u32 daddr,
					    u16 dport, int dif)
{
	struct sock *sk;

	local_bh_disable();
	sk = __nx_tcp_v4_lookup(saddr, sport, daddr, ntohs(dport), dif);
	local_bh_enable();

	return sk;
}
#endif

/*
 *
 */
static struct sock *nx_lookup_tcp_sk(const u32 saddr, const u16 sport,
			      const u32 daddr, const u16 dport,
			      const int dif)
{
	struct sock	*sk;

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,9)
	sk = nx_tcp_v4_lookup(saddr, sport, daddr, dport, dif);
#elif LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,13)
	sk = tcp_v4_lookup(saddr, sport, daddr, dport, dif);
#else
	sk = inet_lookup(&tcp_hashinfo, saddr, sport, daddr, dport, dif);
#endif

        return (sk);
}


#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,14)	
static u32 tcp_v4_synq_hash_tbl(u32 raddr, u16 rport, u32 rnd)
{
	return (jhash_2words(raddr, (u32) rport, rnd) & (TCP_SYNQ_HSIZE - 1));
}
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,14)
static struct open_request *check_tuple_existence(const struct sock *sk,
						  struct open_request ***prevp,
						  __u16 rport, __u32 raddr,
						  __u32 laddr)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	struct tcp_opt		*tp = &(sk->tp_pinfo.af_tcp);
#elif LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,10)
	struct tcp_opt		*tp = tcp_sk(sk);
#endif
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,10)
        struct tcp_listen_opt	*lopt = tp->listen_opt;
#else
	struct tcp_sock		*tp = tcp_sk(sk);
        struct tcp_listen_opt	*lopt = tp->listen_opt;
#endif
	struct open_request	*req, **prev;

	for (prev = &lopt->syn_table[tcp_v4_synq_hash_tbl(raddr, rport,
							  lopt->hash_rnd)];
	    (req = *prev) != NULL; prev = &req->dl_next) {

		if (req->rmt_port == rport &&
		    req->af.v4_req.rmt_addr == raddr &&
		    req->af.v4_req.loc_addr == laddr &&
		    TCP_INET_FAMILY(req->class->family)) {

		    	*prevp = prev;
			break;
		}	
	}	
	return req;
}
#endif

/*
 *
 */
static int nx_lookup_open_req(const struct sock *sk, const __u16 rport,
		       const __u32 raddr, const __u32 laddr)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,14)
        struct request_sock	**prev;
        struct request_sock	*req;
#else
	struct open_request	**prev;
	struct open_request	*req;
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,14)	
	req = inet_csk_search_req(sk, &prev, rport, raddr, laddr);
#else
	req = check_tuple_existence(sk, &prev, rport, raddr, laddr);
#endif
	return (req ? 1 : 0);
}

static int check_non_offloaded_conn(struct sock *sk, u32 saddr, u16 sport,
    u32 daddr)
{
	if (sk->SK_STATE == TCP_LISTEN) {
		return (nx_lookup_open_req(sk, sport, saddr, daddr));
	} else if (TCP_TIME_WAIT == sk->SK_STATE) {
		/* protinfo is not usable here, because the host stack uses it
		   differently. tcp_rcv_state_process -> tcp_time_wait ->
		   __tcp_hashdance will replace sk with a tcp_tw_bucket, and
		   only explicitly initialised fields can be relied upon after
		   that. Luckily, if we are in TIME_WAIT on the host, it could
		   only be because we do not need protinfo! */
#if 0
		printk("%s: Time-wait, not relying on protinfo\n",
		       __FUNCTION__);
		printk("%s: Not offloaded sport = %x, saddr %x, dport = %x, "
		       "daddr = %x\n", __FUNCTION__, th->source,
		       saddr, dport, daddr);
#endif
		return 1;
	} else {
		if (sk->SK_PROTINFO) {
			return 0;
                } else {
			return 1;
		}
        }
}

struct sock *get_offload_listener_sk(struct iphdr *iph)
{
	struct sock	*sk;
	struct tcphdr	*th = (struct tcphdr *)(((char *)iph) + (iph->ihl * 4));
	u32		saddr = iph->saddr, daddr = iph->daddr;
	u16		sport = th->source, dport = th->dest;

#if 0
	printk("%s: SYN saddr:0x%x daddr:0x%x sport:0x%x dport:0x%x\n",
	    __func__, saddr, daddr, sport, dport);
#endif

	if (!(sk = nx_lookup_tcp_sk(saddr, sport, daddr, dport, 0)))
		return(NULL);

	/* lookups do a hold */
	sock_put(sk);

#if 0
	printk("%s: Found sk [%p]\n", __func__, sk);
#endif

	if (check_non_offloaded_conn(sk, saddr, sport, daddr))
		return(NULL);
	return(sk);
}

/*
 * Find the net device for local IP address.
 */
struct net_device *nx_find_dev_for_ip(u32 addr)
{
/* #if (LINUX_VERSION_CODE > KERNEL_VERSION(2,6,13)) && (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)) */
#if	1
	struct net_device	*dev;

	read_lock(&dev_base_lock);
	for (dev = dev_base; dev; dev = dev->next) {
		struct in_device	*indev = dev->ip_ptr;

		if (indev) {
			for_ifa(indev) {
				if (ifa->ifa_address == addr &&
				    netif_running(dev)) {
					dev_hold(dev);
					read_unlock(&dev_base_lock);
					return (dev);
				}
			} endfor_ifa(indev);
		}
	}
	read_unlock(&dev_base_lock);
	return (NULL);
#else
	return (ip_dev_find(addr));
#endif
}

EXPORT_SYMBOL(nx_find_dev_for_ip);
EXPORT_SYMBOL(get_offload_listener_sk);

