/*
 * Copyright (C) 2007 NetXen Inc.
 * 
 * All rights reserved.
 * 
 * The code contained in this file has been developed by NetXen Inc.
 * 
 * By downloading, compiling or using the code in any form, you agree to be bound
 * by the following terms:
 * 
 * 1) You may compile the unmodified source code file and reproduce and install
 * the resulting binary code file solely for your internal use with NetXen Inc.
 * hardware and systems you have purchased.
 * 
 * 2) Except as expressly permitted herein or in a separately signed agreement,
 * any other use, reproduction, distribution, modification, display, creation
 * of derivative works or circulation of the code is strictly prohibited.  You
 * may not transfer the code or assign your rights or obligations hereunder
 * without the prior written consent of NetXen Inc.
 * 
 * 3) This file contains confidential, proprietary and trade secret information
 * of NetXen Inc. and may not be disclosed to any third party or used for any
 * other purpose, except as expressly permitted herein.
 * 
 * 4) THIS SOFTWARE IS PROVIDED “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETXEN INC.
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.  THE APPLICATION OF THE UNITED NATIONS CONVENTION
 * ON CONTRACTS FOR THE INTERNATIONAL SALE OF GOODS IS HEREBY DISCLAIMED.
 * 
 * If you do not agree to the foregoing terms, you may not download, use or
 * compile this code in any form and must promptly delete or destroy all copiesi
 * of the code you have downloaded, compiled or reproduced.
 * 
 * Contact Information:
 * licensing@netxen.com
 * NetXen Inc.
 * 3965 Freedom Circle, Fourth Floor
 * Santa Clara, CA 95054
 * 
 */
/*
 * Copyright and CVS version tags.
 */

/*
 * This has the functions for managing a host table of flows. Currently only
 * TCP.
 */

#include "nx_listen.h"
#include "unm_host_pnic_types.h"

#include <asm/string.h>
#include <linux/slab.h>
#include <linux/errno.h>
#include <linux/vmalloc.h>
#include <linux/jhash.h>
#include <net/ip.h>
#include <net/tcp.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,14)
#include <net/inet_hashtables.h>
#endif


extern index_table_t		g_listener_grp_tbl;

/*
 * This function is used to allocate space for the index table. It allocates the
 * maximum IDs and index pointers. Does not allocate the index itself.
 *
 * Parameters:
 *      None.
 *
 * Return:
 *      0               - If allocated successfully.
 *      ENOMEM          - If out of memory.
 */
int host_indextbl_init(index_table_t *tcp_tbl)
{
	int	i;
	int	alloc_size;

        tcp_tbl->lock = SPIN_LOCK_UNLOCKED;

	alloc_size = sizeof (index_tbl_entry_t) * INDEX_TABLE_SIZE;
	tcp_tbl->p_data = (index_tbl_entry_t *)vmalloc(alloc_size);
	if (tcp_tbl->p_data == NULL) {
		printk("host_indextbl_init: p_data alloc failed\n");
		return (-ENOMEM);
	}

/* 	printk("#####################host_flowtbl_init: 1 = %u\n", */
/* 	       (sizeof (flow_tbl_entry_t) * UNM_HOST_MAX_TCP_CONNECTIONS)); */

	tcp_tbl->free_list.ids = (index_id_t *)kmalloc((sizeof (index_id_t) *
						      FLOW_TABLE_SIZE),
						     GFP_KERNEL);

/* 	printk("#####################host_flowtbl_init: 2 = %u\n", */
/* 	       (sizeof (flow_id_t) * UNM_HOST_MAX_TCP_CONNECTIONS)); */

	if (tcp_tbl->free_list.ids == NULL) {
		printk("host_flowtbl_init: IDs alloc failed\n");
		vfree(tcp_tbl->p_data);
		return (-ENOMEM);
	}


	tcp_tbl->max = INDEX_TABLE_SIZE;
	tcp_tbl->free_list.max = INDEX_TABLE_SIZE;
	tcp_tbl->free_list.tail = INDEX_TABLE_SIZE - 1;


	for (i = 0; i < INDEX_TABLE_SIZE; i++) {
		tcp_tbl->p_data[i].id.idx = UNM_INVALID_INDEX_ID;
		tcp_tbl->p_data[i].data = kmalloc(sizeof (listener_group_t),
						GFP_KERNEL);
		if (!tcp_tbl->p_data[i].data) {
			printk("%s: Unable to allocate data\n", __FUNCTION__);
			goto free_and_out;
		}
		atomic_set(&tcp_tbl->p_data[i].refcnt, 0);

		tcp_tbl->free_list.ids[i].idx = i;

	}

        printk("host_indextbl_init: Initialized\n");

	/* Add everything to the free list */

	return (0);

free_and_out:
	for (i = 0; i < INDEX_TABLE_SIZE; i++) {
		if (tcp_tbl->p_data[i].data)
			kfree(tcp_tbl->p_data[i].data);
	}
	kfree(tcp_tbl->free_list.ids);
	vfree(tcp_tbl->p_data);
	return -ENOMEM;
}

/*
 * This function is used to free space allocated for the flow table.
 *
 * Parameters:
 *      None.
 *
 * Return:
 *      None.
 */
void host_indextbl_destroy(index_table_t *tcp_tbl)
{
	int	i;
	spin_lock(&tcp_tbl->lock);

	for (i = 0; i < INDEX_TABLE_SIZE; i++) {
		if (tcp_tbl->p_data[i].data)
			kfree(tcp_tbl->p_data[i].data);
	}

	spin_unlock(&tcp_tbl->lock);

	kfree(tcp_tbl->free_list.ids);
	vfree(tcp_tbl->p_data);

	return;
}

/*
 * This function is used to allocate a index from the index table.
 * Returns data contained in the index if allocated else returns NULL.
 *
 * Parameters:
 *      index_id: Index of data in host's index-id table
 *
 * Return:
 *      The data that was allocated.
 *      NULL    - If allocation fails.
 */

void *host_indextbl_allocate_index(index_table_t *tcp_tbl, uint32_t *idx)
{
	index_id_t	index_id;
        index_free_ids_t *free_list;
	void		*data = NULL;
	unsigned int 	allocated_index;

	spin_lock_bh(&tcp_tbl->lock);

        free_list = &tcp_tbl->free_list;

	if (free_list->head != free_list->tail) {
		/*
		 * First pick the ID.
		 */
		index_id.idx = allocated_index =
			free_list->ids[free_list->head].idx;

		free_list->head++;
		free_list->head %= free_list->max;

		/*
		 * Now add the data and exit.
		 */
		tcp_tbl->p_data[allocated_index].id.idx = index_id.idx;
		data = tcp_tbl->p_data[allocated_index].data;
		atomic_set(&tcp_tbl->p_data[allocated_index].refcnt, 1);

		tcp_tbl->nr_indexes++;

/*                 printk("%s: id = %u, flow = %p\n", */
/*                        __FUNCTION__, flow_id.id, flow); */
	}

	spin_unlock_bh(&tcp_tbl->lock);

	if (idx) {
		*idx = index_id.idx;
	} else {
		nx_print("index_idx pointer is NULL\n");
	}

	return (data);
}


int host_indextbl_free_data_by_idx(index_table_t *tcp_tbl, uint32_t idx)
{
        index_free_ids_t *free_list;
	uint32_t	old_tail;
	int 		rv = -1;

        free_list = &tcp_tbl->free_list;

	spin_lock_bh(&tcp_tbl->lock);

	if (idx < tcp_tbl->max) {

		if (!atomic_dec_and_test(&tcp_tbl->p_data[idx].refcnt)) {
			printk("%s: Reference not zero\n", __FUNCTION__);
		}

		/*
		 * Clear the table.
		 */
		tcp_tbl->p_data[idx].id.idx = UNM_INVALID_INDEX_ID;

		//tcp_tbl->flows[flow_id.bits.idx].flow = NULL;

		/*
		 * Add to tail of free list.
		 */
		old_tail = tcp_tbl->free_list.tail;
		free_list->tail++;
		free_list->tail %= free_list->max;

		if (free_list->tail != free_list->head) {
			free_list->ids[free_list->tail].idx = idx;
			rv = 0;
		} else {
			printk("host_indextbl_free_tcp_flow: tail reached head "
			       "this should never happen\n");
			tcp_tbl->free_list.tail = old_tail;
		}
		tcp_tbl->nr_indexes--;
	}

	spin_unlock_bh(&tcp_tbl->lock);
	return rv;

}


/*
 * This function is used to get data from the index table.
 *
 * Parameters:
 *      idx             - The index of the data to get.
 *
 * Return:
 *      The data that was found.
 *      NULL            - If there is no data for the index.
 */
void *host_indextbl_get_data_by_idx(index_table_t *tcp_tbl, uint32_t idx)
{
        void	*data = NULL;
        /* printk("%s: id = %u\n", __FUNCTION__, id); */

	spin_lock_bh(&tcp_tbl->lock);

	if (idx < tcp_tbl->max &&
	    tcp_tbl->p_data[idx].id.idx != UNM_INVALID_INDEX_ID) {

		data = tcp_tbl->p_data[idx].data;
		atomic_inc(&tcp_tbl->p_data[idx].refcnt);
	}

	spin_unlock_bh(&tcp_tbl->lock);

//        printk("%s: max=%u idx = %u, tcp_tbl->p_data[idx].id.idx=%u data = %p\n", __FUNCTION__, tcp_tbl->max ,idx, tcp_tbl->p_data[idx].id.idx,data);

	return (data);
}


void host_indextbl_get_reference(index_table_t *tcp_tbl,uint32_t id)
{
        index_id_t	index_id;

	spin_lock_bh(&tcp_tbl->lock);

	index_id.idx = id;

	if (index_id.idx < tcp_tbl->max &&
	    tcp_tbl->p_data[index_id.idx].id.idx == index_id.idx) {

		atomic_inc(&tcp_tbl->p_data[index_id.idx].refcnt);
	}

	spin_unlock_bh(&tcp_tbl->lock);

	return;
}
/*
 * This function is used to release reference to a index from the index table.
 *
 * Parameters:
 *	id		- The ID of the flow to release a reference
 *
 * Return:
 */
void host_indextbl_put_reference(index_table_t *tcp_tbl, uint32_t id)
{
        index_id_t	index_id;

	spin_lock_bh(&tcp_tbl->lock);

	index_id.idx = id;

	if (index_id.idx < tcp_tbl->max &&
	    tcp_tbl->p_data[index_id.idx].id.idx == index_id.idx) {

		atomic_dec(&tcp_tbl->p_data[index_id.idx].refcnt);
	}

	spin_unlock_bh(&tcp_tbl->lock);

	return;
}

int host_indextbl_get_nr_indexes(index_table_t *tcp_tbl)
{
	int cnt;

	spin_lock_bh(&tcp_tbl->lock);
	cnt = tcp_tbl->nr_indexes;
	spin_unlock_bh(&tcp_tbl->lock);

	return cnt;
}
/* Here Starts Listen related function */
static int nx_release_listener_sid(pegnet_sock_id_t *sid)
{
	int	rv = 0;
	
	rv = host_release(sid, 0);
        if (rv != 0) {
		nx_print1("host_release failed %d sid=%p\n", rv,sid);
        }
	return rv;
}

pegnet_sock_id_t *__nx_listener_create(struct net_device *netdev,
    listener_group_t *grp)
{
	pegnet_sock_id_t	*sid;
	uint32_t		grp_idx;
	int			rv;

	/*
	 * Talk to card and get back zero'ed out pegnet_sock_id_t.
	 */
	sid = host_create(netdev, PF_INET, IPPROTO_TCP, &rv);
	if (sid == NULL)
		return ERR_PTR(rv);

	if (grp == NULL) {
		grp = (listener_group_t *)host_indextbl_allocate_index(
		    &g_listener_grp_tbl, &grp_idx);
		if (!grp) {
			host_release(sid, 0);
			return ERR_PTR(-ENOMEM);
		}
		memset(grp, 0, sizeof(listener_group_t));
		spin_lock_init(&grp->lock);
		atomic_set(&grp->child_cnt, 0);
		INIT_LIST_HEAD(&(grp->listener_list));
		grp->index = grp_idx;
	}
	sid->grp = grp;
	list_add_tail(&sid->listener_list, &grp->listener_list);
	return(sid);
}

/*
 * Creates Listener sock
 */
int nx_listener_create(struct net_device *netdev, struct socket *sock,
		       uint8_t rule_flags)
{
        pegnet_sock_id_t        *sid;
        int                     first;
	listener_group_t	*grp;

	first = sock->sk->SK_PROTINFO == NULL ? 1 : 0;
	if (first) {
		grp = NULL;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
        	if (!try_module_get(THIS_MODULE)) {
                	return (-EINVAL);
        	}
#else
        	MOD_INC_USE_COUNT;
#endif
	} else {
		grp = NX_GET_LISTENER_GROUP_FROM_SOCK(sock);
	}

        /* Talk to card and get back sock_id_t */
	sid = __nx_listener_create(netdev, grp);
	if (IS_ERR(sid)) {
		if (first)
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
			module_put(THIS_MODULE);
#else
			MOD_DEC_USE_COUNT;
#endif
		return(PTR_ERR(sid));
	}
	sid->parent = sock;
	sid->zero_copy_enabled = (rule_flags &
				  NETXEN_OFFLOAD_RULE_ZCOPY) ? 1 : 0;
	/* For iscsi initiator there is no listener so don't set the flag. */
//	sid->iscsi = rule_flags & NETXEN_OFFLOAD_RULE_ISCSI;

	if (first) {
		grp = sid->grp;
		grp->parent = sock;
		UNM_SOCK_SET_PEGNET_SOCK_ID(sock, grp);
	}
	return(0);
}

/*
 *
 */
int nx_listen(listener_group_t *grp, int backlog, int do_bind)
{
	int rv = 0;
	struct list_head *node,*_node;
	pegnet_sock_id_t *sid;

	list_for_each_safe(node, _node, &grp->listener_list) {
		sid = list_entry(node, pegnet_sock_id_t, listener_list);
		spin_lock_bh(&sid->lock);
		sid->grp_idx = grp->index;
		spin_unlock_bh(&sid->lock);
		if(0 != (rv = host_listen(sid, backlog, do_bind))) {
			nx_print("Listen Error for Listener grp = %p sid:%p "
				"tnic_adapter=%p dev_port=%d err=%d \n",
				grp,sid,sid->adapter,sid->portnum,rv);
			spin_lock_bh(&grp->lock);
			list_del_init(&sid->listener_list);
			spin_unlock_bh(&grp->lock);
			nx_release_listener_sid(sid);	
		}
	}
	rv = 0;
	if (list_empty(&(grp->listener_list))) {
		rv = -1;
	}
	return rv;
}

int __nx_release_listener_sock(listener_group_t *grp)
{
        struct list_head        *node;
        pegnet_sock_id_t        *sid= NULL;
        int                     rv = 0,err=0;
	
	spin_lock_bh(&grp->lock);
	grp->released = 1;
	while(!list_empty(&(grp->listener_list))) {
		node = grp->listener_list.next;
		sid = list_entry(node, pegnet_sock_id_t,
                                     listener_list);	
		list_del_init(&sid->listener_list);
		spin_unlock_bh(&grp->lock);
		rv = nx_release_listener_sid(sid);
		spin_lock_bh(&grp->lock);
	}
	if (grp->accept_passive.lopt) {
		release_listener_child(grp);
	}
	if(0 == atomic_read(&grp->child_cnt)) {
               	kfree(grp->accept_passive.lopt);
               	grp->accept_passive.lopt = NULL;
		spin_unlock_bh(&grp->lock);
		err = host_indextbl_free_data_by_idx(&g_listener_grp_tbl, grp->index);
        	if (err != 0) {
               		nx_print(": Free from index table failed: %d\n",err);
        	}		
		goto finish;
	}			
	spin_unlock_bh(&grp->lock);
finish:	
        return (rv);
}

int nx_release_listener_sock(struct socket *sock)
{
	int	rv;

	rv = __nx_release_listener_sock(NX_GET_LISTENER_GROUP_FROM_SOCK(sock));
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
        module_put(THIS_MODULE);
#else
        MOD_DEC_USE_COUNT;
#endif
	if (sock->sk != NULL) {
		UNM_SOCK_SET_PEGNET_SOCK_ID(sock, NULL);
	}
	return(rv);
}

/* It Updates connID at host and if delay release bit is on for listener child
*  then send release on it. 
 *
 */
void nx_proc_update_connID(nx_tnic_adapter_t *tnic_adapter, unm_msg_t *msg, 
                        pegnet_sock_id_t *sid, pegnet_inprogress_t *trans)
{
	host_peg_msg_hdr_t      	*hdr = (host_peg_msg_hdr_t *)&(msg->body);
        nx_update_connID_table_t        *resp;
        pegnet_sock_id_t                *syn_sid = NULL;
	listener_group_t        	*grp;
	uint32_t			grp_idx;

        resp = (nx_update_connID_table_t *)&msg->body;
	/* Retrieve the grp */
	grp_idx = host_get_uint32_from_peg(hdr->ctx);
	grp = host_indextbl_get_data_by_idx(&g_listener_grp_tbl, grp_idx);
//	printk("%s: ret_code:%d\n", __func__, resp->ret_code);
	if (!grp) {
		nx_print(":Could not map grp_idx %x to grp\n",grp_idx);
		return;
	}
        /* get the child sid from listener syn q */
        if (!(syn_sid = nx_search_synq(grp, resp->local_ip, resp->local_port,
                                       resp->remote_ip, resp->remote_port))) {
                printk("Could not find child sid for grp [%p]\n", grp);
		host_indextbl_put_reference(&g_listener_grp_tbl, grp_idx);
                return;
        }

        spin_lock_bh(&syn_sid->lock);
	host_assign_qid_from_peg(syn_sid->qid, resp->qid);
        syn_sid->ctx.context =  host_get_uint32_from_peg(resp->child_ctx);
        syn_sid->connID = host_get_uint32_from_peg(resp->child_connID);
        host_update_connID_table(tnic_adapter->p_ftable, syn_sid, resp->child_connID);
	syn_sid->connID_updated = 1;
        if(syn_sid->delay_release == 1) {
		spin_lock_bh(&grp->lock);
                nx_remove_synq(grp, syn_sid);
		spin_unlock_bh(&grp->lock);
                host_release_unaccepted_child(syn_sid);
                syn_sid->release_seen = 1;
        }
        spin_unlock_bh(&syn_sid->lock);
	host_indextbl_put_reference(&g_listener_grp_tbl, grp_idx);
        return;
}

/* Handle handshake complete msg. 
 *
 */
void nx_proc_tcp_handshake_response(nx_tnic_adapter_t *tnic_adapter,
					    unm_msg_t *msg,
					    pegnet_sock_id_t *sid,
					    pegnet_inprogress_t *trans)
{
	host_peg_msg_hdr_t      	*hdr = (host_peg_msg_hdr_t *)&(msg->body);
	tcp_handshake_complete_t	*resp;
	pegnet_sock_id_t		*syn_sid = NULL;
	listener_group_t        	*grp;
	uint32_t			grp_idx;

	resp = (tcp_handshake_complete_t *)&msg->body;
	/* Retrieve the grp */
	grp_idx = host_get_uint32_from_peg(hdr->ctx);
	grp = host_indextbl_get_data_by_idx(&g_listener_grp_tbl, grp_idx);
//	printk("%s: ret_code:%d\n", __func__, resp->ret_code);
	if (!grp) {
		nx_print(":Could not map grp_idx %x to grp\n",grp_idx);
		return;
	}
	spin_lock_bh(&grp->lock);

	if (grp->released) {
		UNM_HOST_PNIC_STATS_INC(DBG_CNTR_CONN_FOR_FREED_LISTENER);
		spin_unlock_bh(&grp->lock);
		host_indextbl_put_reference(&g_listener_grp_tbl, grp_idx);
		return;
	}

	/* get the child sid from listener syn q */
	syn_sid = nx_search_synq(grp, resp->local_ip, resp->local_port,
				 resp->remote_ip, resp->remote_port);
	if (!syn_sid) {
		printk("Could not find child sid for sid [%p]\n", sid);
		spin_unlock_bh(&grp->lock);
		host_indextbl_put_reference(&g_listener_grp_tbl, grp_idx);
		return;
	}

	/* remove this from the syn queue */
	nx_remove_synq(grp, syn_sid);
	spin_lock_bh(&syn_sid->lock);
	syn_sid->accept_active.synq_flag = 0;
	spin_unlock_bh(&syn_sid->lock);

	/* free the sid if we got error */
	if (resp->ret_code != 0) {
		spin_unlock_bh(&grp->lock);

		UNM_HOST_PNIC_STATS_INC(DBG_CNTR_ERR_PASSIVE_CONN_COMPLETE);
		if (resp->ret_code != PEGNET_SYN_OFFLOAD_RESP_LISTENER_FAIL) {
			/* processing has gone past the listening peg,
			   we should save some fields for cleanup. */
			host_assign_qid_from_peg(syn_sid->qid, resp->qid);
			host_release_unaccepted_child(syn_sid);
			/*
			 * There is no way a sock cleanup can come before this
			 * message. If it comes in before this or is processed
			 * before this we have a serious problem.
			 *
			 * No need to free the sid here. The sid will be freed
			 * when the sock cleanup comes down.
			 */
			syn_sid->release_seen = 1;
		} else {
			free_pegnet_sock_id(syn_sid);
		}
		host_indextbl_put_reference(&g_listener_grp_tbl, grp_idx);
		return;
	}

	spin_unlock_bh(&grp->lock);
        spin_lock_bh(&syn_sid->lock);
	/* make the child sid usable */
	host_assign_qid_from_peg(syn_sid->qid, resp->qid);
	host_assign_qid_from_peg(syn_sid->rcvBuf_qid, resp->rcvBuf_qid);
	syn_sid->ctx.context =  host_get_uint32_from_peg(resp->child_ctx);
	syn_sid->connID = host_get_uint32_from_peg(resp->child_connID);
	host_update_connID_table(tnic_adapter->p_ftable, syn_sid,
				 resp->child_connID);
        spin_unlock_bh(&syn_sid->lock);

	spin_lock_bh(&grp->lock);
	/* add to the listener accept q */
	list_add_tail(&syn_sid->accept_active.list,
		      &grp->accept_passive.accept_queue);

	grp->accept_passive.accepted_len++;
	spin_unlock_bh(&grp->lock);

//	printk("%s: About to wake up sk=%p sid=%p grp=%p\n", __func__, sid->parent->sk,sid,grp);

#ifndef RDMA_MODULE
	/* wakeup any pending accepts */
	if (grp->parent->sk) {
		bh_lock_sock(grp->parent->sk);
		if (grp->parent->sk->SK_DATA_READY) {
			grp->parent->sk->SK_DATA_READY(grp->parent->sk, 0);
		}
		bh_unlock_sock(grp->parent->sk);
	}
#endif /* !RDMA_MODULE */
	host_indextbl_put_reference(&g_listener_grp_tbl, grp_idx);
	return;
}


/*
 * A child sk has declared that it can be pruned
 */
void nx_proc_prunable_sk(nx_tnic_adapter_t *tnic_adapter,
				 unm_msg_t *msg, pegnet_sock_id_t *sid,
				 pegnet_inprogress_t *trans)
{
	host_peg_msg_hdr_t      *hdr = (host_peg_msg_hdr_t *)&(msg->body);
	pegnet_sock_id_t	*child;
	tcp_prunable_sk_t	*pr_sk = (tcp_prunable_sk_t *)&msg->body;
        nx_lopt_t		*lopt;
        listener_group_t	*grp;
	uint32_t		grp_idx;

	/* Retrieve the grp */
	grp_idx = host_get_uint32_from_peg(hdr->ctx);
	grp = host_indextbl_get_data_by_idx(&g_listener_grp_tbl, grp_idx);
	if (!grp) {
		nx_print(":Could not map grp_idx %x to grp\n",grp_idx);
		return;
	}
	spin_lock_bh(&grp->lock);
	if (!(child = nx_search_synq(grp, pr_sk->local_ip,
					pr_sk->local_port,
					pr_sk->remote_ip,
					pr_sk->remote_port)) ) {
		nx_print("Could not find child of %p: rport = %x, rip = %x\n",
			 grp, pr_sk->remote_port, pr_sk->remote_ip);
		spin_unlock_bh(&grp->lock);
		host_indextbl_put_reference(&g_listener_grp_tbl, grp_idx);
		return;
	}

	spin_unlock_bh(&grp->lock);

        spin_lock_bh(&child->lock);
	host_assign_qid_from_peg(child->qid, pr_sk->child_qid);
	child->ctx.context =  host_get_uint32_from_peg(pr_sk->child_ctx);
	child->connID = host_get_uint32_from_peg(pr_sk->child_connID);
	host_update_connID_table(tnic_adapter->p_ftable, child,
				 pr_sk->child_connID);

        spin_unlock_bh(&child->lock);

	spin_lock_bh(&grp->lock);
	lopt = grp->accept_passive.lopt;
	list_add_tail(&child->accept_active.young_sid, &lopt->youngsters);
	spin_unlock_bh(&grp->lock);

	UNM_HOST_PNIC_STATS_INC(DBG_CNTR_TOTAL_PRUNABLES);

	host_indextbl_put_reference(&g_listener_grp_tbl, grp_idx);
	return;
}


/*
 * It search for child sid inside synq table.  
 */
pegnet_sock_id_t *nx_search_synq(listener_group_t *grp,
					uint32_t local_ip,
					uint16_t local_port,
					uint32_t remote_ip,
					uint16_t remote_port)
{
	u32 h;
	struct list_head *node;
	pegnet_sock_id_t *tmp;
        nx_lopt_t *lopt = grp->accept_passive.lopt;

	if (!lopt) {
		printk("%s: lopt NULL for grp [%p]\n", __FUNCTION__, grp);
	       	return 0;
	}

       	h = jhash_2words(remote_ip, remote_port, lopt->hash_rnd)
				& (NX_SYNQ_HSIZE - 1);

	local_port = htons(local_port);

	list_for_each(node, &lopt->syn_table[h]) {
		tmp = list_entry(node, pegnet_sock_id_t,
					accept_active.list);

		if (tmp->rip_addr.v4 == remote_ip &&
				tmp->rport == remote_port &&
				tmp->lip_addr.v4 == local_ip &&
				tmp->lport == local_port) {
			return tmp;
		}
	}

	return NULL;
}

/* perform some sanity checks on the packet
 */
static int check_syn_packet(struct sk_buff *skb, struct iphdr *iph)
{
	struct tcphdr	*th;
	uint32_t	len;

	/* ip version
	 * hdr length
	 * total length
	 */
	if (iph->ihl < 5 || iph->version != 4 ) {
		return 1;
	}

	len = ntohs(iph->tot_len);
	if (skb->len < len || len < (iph->ihl << 2)) {
		return 1;
	}
	
	/* tcp doff
	 */
	th = (struct tcphdr *)(((char *)iph) + (iph->ihl * 4));
	if (th->doff < sizeof(struct tcphdr) / 4) {
		return 1;
	}

	return 0;

}

/* check -
 *   	- ip options
 *	- syn with data
 *	- tcp options
 */
static int check_syn_options(struct sk_buff *skb, struct iphdr *iph)
{
	struct tcphdr	*th;

	th = (struct tcphdr *)(((char *)iph) + (iph->ihl * 4));

	// printk("%s: iph->ihl:%d so(iphdr):%d tot_len:%d th->doff:%d\n", __func__, iph->ihl, sizeof(struct iphdr), ntohs(iph->tot_len), th->doff);
	/* no ip options */
	if ( (iph->ihl * 4) > sizeof(struct iphdr))
		return 1;

	/* no syn with data */
	if ( ntohs(iph->tot_len) > ( (iph->ihl << 2) + (th->doff << 2)) )
		return 1;

	/* Will the TCP header fit it a single host-to-card message */
	if (NX_MAX_TCP_OFFLOAD_HDR_SIZE < (th->doff << 2)) {
		printk("%s: DOFF = %d is too big (max = %d)\n", __FUNCTION__,
				(int)th->doff, (int)NX_MAX_TCP_OFFLOAD_HDR_SIZE);
		return 1;
	}

	return 0;
}

/* offload the syn -
 *   - allocate sid
 *   - add sid to synq
 *   - send UNM_MSGTYPE_RXED_SYN_OFFLOAD to card
 */
static int offload_syn(struct net_device *netdev, pegnet_sock_id_t *parent_sid, 
		struct sk_buff *skb, struct iphdr *iph, unsigned short backlog)
{
	pegnet_sock_id_t	*sid;
	listener_group_t        *grp;
	unm_msg_t		msg[NX_NUM_MSGS_PER_SYN];
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,10)	
	struct tcp_options_received		opt_rx = {0};
#else
	struct tcp_opt		opt_rx = {0};
#endif
	nx_lopt_t		*lopt;
	uint32_t		hash;
	struct list_head	*node;
	struct rtable		*rt;
	int			ifport;
	nx_tnic_adapter_t *tnic_adapter = nx_get_dev_tnic_adapter(netdev);

        grp = NX_GET_LISTENER_GROUP_FROM_SID(parent_sid);
        lopt = grp->accept_passive.lopt;
	skb->h.th = (struct tcphdr *)(((char *)iph) + (iph->ihl * 4));

	/*
	 * Check if it is a retransmitted SYN.
	 */
	hash = jhash_2words(iph->saddr, skb->h.th->source, lopt->hash_rnd);
	hash &= (NX_SYNQ_HSIZE - 1);

	list_for_each(node, &lopt->syn_table[hash]) {
		sid = list_entry(node, pegnet_sock_id_t, accept_active.list);
		if (sid->rip_addr.v4 == iph->saddr &&
		    sid->lip_addr.v4 == iph->daddr &&
		    sid->rport == skb->h.th->source) {
			printk("%s: Dropping retransmitted SYN. Should "
			       "forward.\n", __FUNCTION__);
			return (0);
		}
	}

	/* ip options and syn with data not allowed */
	if (check_syn_options(skb, iph)) {
		printk("%s: IP options packet forwarded\n", __FUNCTION__);
		return (1);
	}

	/* card resources */
	if (check_card_resources(tnic_adapter->p_ftable)) {
		static unsigned int print_cnt = 0;
		if ((print_cnt % 10000) == 0) {
			nx_print("Num connections maxed out on card: %u\n", print_cnt);
		}
		print_cnt++;

		if (parent_sid->rdmacookie != NULL)
			return(0);
		return (1);
	}

	/*
	 * Check if we can accept it.
	 */
	if ((lopt->qlen > NX_MAX_SYNQ_LEN) ||
	    grp->accept_passive.accepted_len > backlog) {

		UNM_HOST_PNIC_STATS_INC(DBG_CNTR_SYNQ_DROPPED);
		return (0);
	}

	/* We will accept this one, but do we need pruning? */
	if (lopt->qlen >= (NX_MAX_SYNQ_LEN >> 1)) {
		nx_prune_synq(grp, NX_SYNQ_PRUNETYPE_BH);
	}

	sid = alloc_pegnet_sock_id(netdev);
	if (!sid) {
		printk("null sid\n");
		return (1);
	}

	/* parse tcp options */
	tcp_parse_options(skb, &opt_rx, 0);

	/*
	 * Do the output routing and if required populate the card with the
	 * route.
	 */
	if (nx_route_connection(&rt, iph->daddr, skb->h.th->dest, iph->saddr,
				skb->h.th->source, iph->tos,
				0/* dev->ifindex */)) {
		nx_print("Unable to route connection\n");
		free_pegnet_sock_id(sid);
		return (1);
	}

	ifport = nx_get_device_port(NX_KC_ROUTE_DEVICE(rt));
	if (ifport < 0) {
		nx_print("Not netxen device? Weird\n");
		free_pegnet_sock_id(sid);
		ip_rt_put(rt);
		return (1);
	}

	if (nx_route_add(tnic_adapter, rt, ifport)) {
		nx_print("Error: Route add failed\n");
		free_pegnet_sock_id(sid);
		ip_rt_put(rt);
		return (1);
	}
	ip_rt_put(rt);
	nx_set_route_reference(sid, iph->daddr, skb->h.th->dest, iph->saddr,
			       skb->h.th->source);


	/* initialize some fields */
	sid->lport = skb->h.th->dest;
	sid->rport = skb->h.th->source;
	sid->lip_addr.v4 = iph->daddr;
	sid->rip_addr.v4 = iph->saddr;
	if (parent_sid->rdmacookie != NULL) {
		extern void *(*iw_con_handler)(void *cbarg, void *childcookie,
		    u32 locip, u32 remip, u16 lport, u16 rport);
		void	*info;
		info = iw_con_handler(parent_sid->rdmacookie, sid,
		    sid->lip_addr.v4, sid->rip_addr.v4, sid->lport, sid->rport);
		if (info == NULL) {
			free_pegnet_sock_id(sid);
			return (1);
		}
		sid->rdmacookie = info;
	}
	sid->zero_copy_enabled = parent_sid->zero_copy_enabled;
	sid->accept_active.parent_sid = parent_sid;
	sid->accept_active.synq_flag = 1;
	sid->grp = grp;

	/* add to the syn q */
//	printk("%s: raddr:0x%x rport:0x%x rnd:0x%x hash:%d\n", __FUNCTION__, sid->rip_addr.v4, sid->rport, lopt->hash_rnd, h);

	list_add_tail(&sid->accept_active.list, &lopt->syn_table[hash]);
	lopt->qlen++;

	INIT_LIST_HEAD(&sid->accept_active.young_sid); /* not needed */

	/* First, copy out the iphdr */
	UNM_MSGQ_HDR(msg[NX_SYN_IPH_IDX].hdr, parent_sid->qid.major,
			parent_sid->qid.minor,
			RXED_SYN_OFFLOAD);
	memcpy(&msg[NX_SYN_IPH_IDX].body, iph, iph->ihl << 2);

	/* Next, the tcp hdr */
	UNM_MSGQ_HDR(msg[NX_SYN_TCPH_IDX].hdr, parent_sid->qid.major,
			parent_sid->qid.minor,
			RXED_SYN_OFFLOAD);
	memcpy(&msg[NX_SYN_TCPH_IDX].body, skb->h.th, skb->h.th->doff << 2);

#if 0
	printk("%s: lip:0x%x rip:0x%x lport:0x%x rport:0x%x conn:0x%x\n",
	       __func__, syn_req->local_ip.v4, syn_req->remote_ip.v4,
	       syn_req->local_port, syn_req->remote_port, syn_req->connID);
#endif
	/* here we go */
	if (unm_send_h2p_msg_atomic(netdev, &msg[0], NX_NUM_MSGS_PER_SYN)) {
		printk("error sending compressed syn\n");
		nx_remove_synq(grp, sid);
		free_pegnet_sock_id(sid);
		if (parent_sid->rdmacookie) {
			kfree(sid->rdmacookie);
			sid->rdmacookie = NULL;
		}
	}

	return (0);
}

#ifndef RDMA_MODULE

/*
 * NULL sid returned implies SYN should not be offload processed, and *rvp
 * indicates whether nic should send SYN up to host or not.
 */
static pegnet_sock_id_t *get_offload_listener(nx_tnic_adapter_t *tnic_adapter,
    u16 lid, struct iphdr *iph, unsigned short *blogp, int *rvp)
{
	struct sock		*sk;
	pegnet_sock_id_t	*sid;
	listener_group_t	*grp;

	if ((sk = get_offload_listener_sk(iph)) == NULL) {
		*rvp = 1;
		return(NULL);
	}

	*blogp = sk->sk_max_ack_backlog;
	grp = sk->SK_PROTINFO;
	if (grp == NULL) {
		printk("%s: Listener group not found for sk [%p]\n",
		    __FUNCTION__, sk);
		*rvp = 1;
		return(NULL);
	}
	if (grp->released) {
		UNM_HOST_PNIC_STATS_INC(DBG_CNTR_SYN_FOR_FREED_LISTENER);
		*rvp = 0;
		return(NULL);
	}
	sid = get_sid_by_grp(tnic_adapter, grp);
	if (sid == NULL) {
		printk("%s: sid not found for sk [%p] listener group [%p] adapter [%p]\n", __FUNCTION__, sk, grp, tnic_adapter);
		*rvp = 1;
		return(NULL);
	}

	host_flowtbl_get_reference(tnic_adapter->p_ftable, sid->connID);
	return(sid);
}

#else /* !RDMA_MODULE */

/*
 * NULL sid returned implies SYN should not be offload processed, and *rvp
 * indicates whether nic should send SYN up to host or not.
 */
static pegnet_sock_id_t *get_offload_listener(nx_tnic_adapter_t *tnic_adapter,
    u16 lid, struct iphdr *iph, unsigned short *blogp, int *rvp)
{
	pegnet_sock_id_t	*sid;
	listener_group_t	*grp;

	sid = (pegnet_sock_id_t *)host_flowtbl_get_flow_by_idx(
	    tnic_adapter->p_ftable, lid);
	grp = sid->grp;
	*blogp = 5;
	if (grp->released) {
		host_flowtbl_put_reference(tnic_adapter->p_ftable, sid->connID);
		UNM_HOST_PNIC_STATS_INC(DBG_CNTR_SYN_FOR_FREED_LISTENER);
		*rvp = 0;
		return(NULL);
	}
	return(sid);
}

#endif /* !RDMA_MODULE */

/* This will handle SYNs for offloaded connections
 *
 * Return - 0: pegnet handled the SYN, nic should drop it
 *          1: nic should handle it
 */
int nx_handle_offloaded_syn(struct net_device *netdev,
			    uint16_t listener_id, struct sk_buff *skb)
{
	struct iphdr		*iph;
	struct tcphdr		*th;
	pegnet_sock_id_t	*sid;
	int			rv;
	unsigned short		backlog;
	nx_tnic_adapter_t	*tnic_adapter = nx_get_dev_tnic_adapter(netdev);

	if (!skb || !skb->data) {
		printk("skb or skb data\n");
		return 1;
	}

	if (skb->protocol == htons(ETH_P_8021Q)) {
		iph = (struct iphdr *)(skb->data + VLAN_HLEN);
	} else {
		iph = (struct iphdr *)skb->data;
	}
	th = (struct tcphdr *)(((char *)iph) + (iph->ihl * 4));
    
	if (!(tcp_flag_word(th) & TCP_FLAG_SYN) ||
	    (tcp_flag_word(th) & (TCP_FLAG_CWR | TCP_FLAG_ECE | TCP_FLAG_URG |
				  TCP_FLAG_ACK | TCP_FLAG_RST |
				  TCP_FLAG_FIN))) {
		printk("%s: Not a packet that can be processed\n",
		       __FUNCTION__);
		return (1);
	}

	sid = get_offload_listener(tnic_adapter, listener_id, iph, &backlog,
	    &rv);
	if (sid == NULL)
		return(rv);

	rv = 0;
	spin_lock_bh(&sid->lock);
	if (!sid->listener) {
		printk("%s: sid for listener_id[%u] is not a listener\n",
		       __FUNCTION__, listener_id);
		rv = 1;
		goto done;
	}

	if (sid->released) {
		UNM_HOST_PNIC_STATS_INC(DBG_CNTR_SYN_FOR_FREED_LISTENER);
		goto done;
	}

	/* do some sanity tests on the packet */
	if (check_syn_packet(skb, iph)) {
		printk("%s: Invalid packet, drop it\n", __FUNCTION__);
		goto done;
	}

	/* Ok, everything's fine. Let's offload this */
	if (offload_syn(netdev, sid, skb, iph, backlog)) {
		rv = 1;
		goto done;
	}

  done:
	spin_unlock_bh(&sid->lock);

	host_flowtbl_put_reference(tnic_adapter->p_ftable, sid->connID);
	return (rv);
}


/*** SYN-queue pruning implementation ***/

#define MAX_SIDS_PER_PRUNING_CYCLE 8
#define MAX_PRUNED_SIDS_BH         8
#define MAX_PRUNED_SIDS            64

static nx_kc_work_queue_t	g_synq_pruning_task;
static nx_synq_prun_requests_t g_prune_reqs;



static void nx_synq_pruning_worker(void *request_q);
static void nx_schedule_pruning(listener_group_t  *grp);
static listener_group_t *nx_get_pruning_request(nx_synq_prun_requests_t *q);

void nx_init_pruning_task(void)
{
	INIT_LIST_HEAD(&g_prune_reqs.sid_list);
	spin_lock_init(&g_prune_reqs.lock);

	NX_KC_INIT_WORK(&g_synq_pruning_task,
			nx_synq_pruning_worker, &g_prune_reqs);
	return;
}

/* sid spin lock should be held when we enter this function */
void nx_prune_synq(listener_group_t   *grp, unsigned int type)
{
	unm_msg_t                       msg;
	nx_send_immediate_conn_compl_t  *req =
		(nx_send_immediate_conn_compl_t *)&msg.body;

	unsigned int			i = 0;
	unsigned int			max = ((type == NX_SYNQ_PRUNETYPE_BH) ?
					       MAX_PRUNED_SIDS_BH :
					       MAX_PRUNED_SIDS);

	pegnet_sock_id_t                *child;
	struct list_head                *node = NULL, *tmp = NULL;
        nx_lopt_t			*lopt = grp->accept_passive.lopt;


	if (NX_SYNQ_PRUNETYPE_BH == type) {
		if (grp->accept_passive.pruning_active) {
			return;
		}
	}

	req->hdr.opcode = UNM_H2P_OPCODE_FORCE_3WAY_HSHAKE_COMPL;

	list_for_each_safe(node, tmp, &lopt->youngsters) {
		child = list_entry(node, pegnet_sock_id_t,
				accept_active.young_sid);

		UNM_MSGQ_HDR(msg.hdr,
			     child->accept_active.parent_sid->qid.major,
			     child->accept_active.parent_sid->qid.minor,
			     HOST_REQUEST);
		msg.hdr.dst_major 	= child->qid.major;
		msg.hdr.dst_minor 	= child->qid.minor;

		req->local_ip.v4        = child->lip_addr.v4;
		req->remote_ip.v4       = child->rip_addr.v4;
		req->local_port         = child->lport;
		req->remote_port        = child->rport;

		/* You can be young only once! */
		list_del_init(&child->accept_active.young_sid);

		if (unm_send_h2p_msg(NX_GET_DEV_FROM_TNIC_ADAPTER(child->adapter, child->portnum),
				     &msg, sizeof(*req))) {
			nx_print("Unable to send pruning request\n");
		} else {
			UNM_HOST_PNIC_STATS_INC(DBG_CNTR_TOTAL_PRUNED);
		}

		if (++i == max) {
			break;
		}
	}

	if (NX_SYNQ_PRUNETYPE_BH == type) {
                grp->accept_passive.pruning_active = 1;
                nx_schedule_pruning(grp);
	}
	else {
		grp->accept_passive.pruning_active = 0;
	}

	return;
}

static void nx_schedule_pruning(listener_group_t  *grp)
{
	/* Update the task queue and request scheduling */

	/* this should not be needed here */
	INIT_LIST_HEAD(&grp->accept_passive.pruning);

	spin_lock(&g_prune_reqs.lock);
        list_add_tail(&grp->accept_passive.pruning,
                        &g_prune_reqs.sid_list);
        spin_unlock(&g_prune_reqs.lock);
        grp->accept_passive.pruning_active = 1;

	NX_KC_SCHEDULE_WORK(&g_synq_pruning_task);
	return;
}

static void nx_synq_pruning_worker(void *request_q)
{
	nx_synq_prun_requests_t *q = &g_prune_reqs;
        listener_group_t *next_grp = NULL;
        unsigned int    i = MAX_SIDS_PER_PRUNING_CYCLE;

        while (i-- && (next_grp = nx_get_pruning_request(q))) {
                if (next_grp->released) {
                        i++;
                        continue;
                }

                spin_lock_bh(&next_grp->lock);
                nx_prune_synq(next_grp, NX_SYNQ_PRUNETYPE_NORMAL);
                spin_unlock_bh(&next_grp->lock);
        }

	spin_lock_bh(&q->lock);
	if (!list_empty(&q->sid_list)) {
		spin_unlock_bh(&q->lock);
		NX_KC_SCHEDULE_WORK(&g_synq_pruning_task);
	} else {
		spin_unlock_bh(&q->lock);
	}

	return;
}

listener_group_t *nx_get_pruning_request(nx_synq_prun_requests_t *q)
{
        listener_group_t *grp = NULL;
        struct list_head *node;

        spin_lock_bh(&q->lock);

        list_for_each(node, &q->sid_list) {
                grp = list_entry(node, listener_group_t,
                                accept_passive.pruning);

                list_del_init(&grp->accept_passive.pruning);
                break;
        }

        spin_unlock_bh(&q->lock);

        return grp;

}

void nx_remove_synq(listener_group_t *grp, pegnet_sock_id_t *sid)
{
	list_del(&sid->accept_active.list);

	list_del_init(&sid->accept_active.young_sid); /* We may be attached */
	grp->accept_passive.lopt->qlen--;
}
/*
 *
 */
static int nx_release_accept_queue_sid(pegnet_sock_id_t *sid)
{
	int			rv = 0;
	pegnet_queue_addr_t	remote = sid->qid;
	release_request_t	*host_req;
	unm_msg_t               msg;
	int			flag = 0;

	if (null_queue(remote)) {
		nx_print("null_queue\n");
		rv  = -ENOTSOCK;
		goto done;
	}

	spin_lock_bh(&sid->lock);

	sid->released = 1;

	spin_unlock_bh(&sid->lock);



	host_req = (release_request_t *)&msg.body;
        memset(host_req, 0, sizeof (*host_req));
	host_req->force = 1;
	host_req->reset = 1;

	host_req->hdr.ctx = sid->ctx.context;
	host_req->listener_flag = 1;

	UNM_HOST_PNIC_STATS_INC(DBG_CNTR_FREED_FROM_ACCEPTQ);

	UNM_MSGQ_HDR(msg.hdr, sid->qid.major, sid->qid.minor, RELEASE_REQUEST);
	rv = unm_send_h2p_msg_sleep(NX_GET_DEV_FROM_TNIC_ADAPTER(sid->adapter, sid->portnum), &msg, 1);

	if (rv == 0) {
	} else {
		rv = -EAGAIN;
		nx_print("Could not send message\n");
	}

	spin_lock_bh(&sid->lock);
	/*
	 * cleanup_flag == 1. This means the sock cleanup message has come in
	 *		      from the card. In this case we can free
	 *		      the whole socket normally since the card
	 *		      has released it.
	 * cleanup_flag == 2.  This means the sid is queued in the to
	 *		      be killed list. This state can not happen
	 *		      here since release_seen is not set yet.
	 */
	if (sid->cleanup_flag == 0) {
		sid->release_seen = 1;
		sid->op_needed |= CLEANUP_GLOBAL_RX_BUF;
	} else if (sid->cleanup_flag == 1) {
		flag = 1;
	} else {
		nx_print("%s: cleanup_flag=%u, sid:%p. INVALID STATE\n",
		       __FUNCTION__, sid->cleanup_flag, sid);
	}
	spin_unlock_bh(&sid->lock);
	if (flag == 1) {
		cleanup_rxbuffers_global(sid);
		free_pegnet_sock_id(sid);
	}

  done:
	return (rv);
} /* nx_release_accept_queue_sid */


/* clear out the syn q
 */
static void host_release_synq(listener_group_t *grp, pegnet_sock_id_t *sid)
{
        unm_msg_t               msg;
        release_request_t       *host_req;
        int                     rv = 0;
        int                     flag = 0;



	atomic_inc(&grp->child_cnt);
        spin_lock_bh(&sid->lock);
        sid->released = 1;

        if(sid->connID_updated == 0) {
                sid->delay_release = 1;
                spin_unlock_bh(&sid->lock);
                return;
        }
        spin_unlock_bh(&sid->lock);

        /* remove this from the syn queue */
        nx_remove_synq(grp, sid);

        host_req = (release_request_t *)&msg.body;
        memset(host_req, 0, sizeof (*host_req));
        host_req->force = 1;
        host_req->reset = 1;

        host_req->hdr.ctx = sid->ctx.context;
	host_req->listener_flag = 1;

        UNM_HOST_PNIC_STATS_INC(DBG_CNTR_FREED_FROM_ACCEPTQ);

        UNM_MSGQ_HDR(msg.hdr, sid->qid.major, sid->qid.minor, RELEASE_REQUEST);
        rv = unm_send_h2p_msg_sleep(NX_GET_DEV_FROM_TNIC_ADAPTER(sid->adapter, sid->portnum), &msg, 1);

        if (rv == 0) {
        } else {
                rv = -EAGAIN;
                nx_print("Could not send message\n");
        }

        spin_lock_bh(&sid->lock);
        /*
         * cleanup_flag == 1. This means the sock cleanup message has come in
         *                    from the card. In this case we can free
         *                    the whole socket normally since the card
         *                    has released it.
         * cleanup_flag == 2.  This means the sid is queued in the to
         *                    be killed list. This state can not happen
         *                    here since release_seen is not set yet.
         */
        if (sid->cleanup_flag == 0) {
                sid->release_seen = 1;
		sid->op_needed |= CLEANUP_GLOBAL_RX_BUF;
        } else if (sid->cleanup_flag == 1) {
                flag = 1;
        } else {
                nx_print("%s: cleanup_flag=%u, sid:%p. INVALID STATE\n",
                       __FUNCTION__, sid->cleanup_flag, sid);
        }
        spin_unlock_bh(&sid->lock);
        if (flag == 1) {
        	cleanup_rxbuffers_global(sid);
                free_pegnet_sock_id(sid);
        }
        return;
}
/*
 * This function frees up all the resources associated with a listen socket
 * queues.
 *
 * Parameters:
 *	sock		- The listening socket for which the resource is to be
 *			  freed.
 * Return:
 *	None.
 */
void release_listener_child(listener_group_t *grp)
{
	pegnet_sock_id_t	*newsid;
	struct list_head 	*node, *_node;
	int			i;

	/*
	 * For the accepted ones we need to call host_release so peg can
	 * free the resources.
	 */
	while (!list_empty(&grp->accept_passive.accept_queue)) {
		newsid = list_entry(grp->accept_passive.accept_queue.next,
				     pegnet_sock_id_t, accept_active.list);
                list_del(&newsid->accept_active.list);
                grp->accept_passive.accepted_len--;
		nx_release_accept_queue_sid(newsid);
	}

	/* clear out the syn queue */
	for (i = 0; i < NX_SYNQ_HSIZE && grp->accept_passive.lopt; i++) {
		list_for_each_safe(node, _node,
				   &grp->accept_passive.lopt->syn_table[i]) {
			newsid = list_entry(node, pegnet_sock_id_t,
					    accept_active.list);

                        host_release_synq(grp, newsid);

		}
	}

}

/*
 *
 */
int unm_host_is_accept_queue_empty(struct socket *sock)
{
        //pegnet_sock_id_t        *sid;
        listener_group_t        *grp;
        int                     ret;

        grp = (listener_group_t *) UNM_SOCK_GET_PEGNET_SOCK_ID(sock);

        spin_lock_bh(&grp->lock);
        ret = list_empty(&(grp->accept_passive.accept_queue));
        spin_unlock_bh(&grp->lock);

        return (ret);
}

/*
 *
 */
int unm_host_get_next_accept_stack(struct socket *sock)
{
        listener_group_t        *grp;
        int                     ret;

        grp = (listener_group_t *) NX_GET_LISTENER_GROUP_FROM_SOCK(sock);
        spin_lock_bh(&grp->lock);
        grp->accept_passive.nxt_stack++;
        grp->accept_passive.nxt_stack &= UNM_ACCEPT_STACK_MASK;
        ret = grp->accept_passive.nxt_stack;
        spin_unlock_bh(&grp->lock);

        return (ret);
}

/* Finds matching sid for given TOE adapter 
 *
 */
pegnet_sock_id_t *get_sid_by_grp(nx_tnic_adapter_t *tnic_adapter, listener_group_t *listener_grp)
{
	struct list_head *node;
	pegnet_sock_id_t *sid;
	list_for_each(node, &listener_grp->listener_list) {
		sid = list_entry(node, pegnet_sock_id_t, listener_list);
		if (sid->adapter == tnic_adapter)
			return sid;
	}
	return NULL;
}

/* Quietly clean up the child which was lying in the listener's
   accept queue. */
void host_release_unaccepted_child(pegnet_sock_id_t *sid)
{
	unm_msg_t		msg;
	tcp_child_release_t	*rel = (tcp_child_release_t *)&msg;

	UNM_MSGQ_HDR(msg.hdr, sid->qid.major, sid->qid.minor, CHILD_RELEASE);

	rel->inform_host = 0;

	if (unm_send_h2p_msg(NX_GET_DEV_FROM_TNIC_ADAPTER(sid->adapter,
							  sid->portnum),
			     &msg, sizeof(tcp_child_kill_t))) {
		nx_print("Error sending msg\n");
	}

	return;
}

/*
 * This is a non blocking accept. If there is any entry in the accepted queue
 * it returns it. It also posts more Accept Requests to the card if necessary.
 */
pegnet_sock_id_t *unm_host_async_accept(listener_group_t *grp, int flags,
					int *err)
{
        UNUSED static char      *me = "unm_host_async_accept";
        pegnet_sock_id_t        *newsock;

        *err = 0;

        spin_lock_bh(&grp->lock);

        if (!list_empty(&grp->accept_passive.accept_queue)) {
                newsock = list_entry(grp->accept_passive.accept_queue.next,
                                     pegnet_sock_id_t, accept_active.list);
                list_del(&newsock->accept_active.list);
                grp->accept_passive.accepted_len--;

                /* We do not need a reference to our parent sid any more.
                   Do it with lock_bh help to avoid race with an RST event
                   reaching an ESTD child.  */
                newsock->accept_active.parent_sid = NULL;
		newsock->grp = NULL;
        } else {
                newsock = NULL;
                *err = -1;
        }

        spin_unlock_bh(&grp->lock);

        return (newsock);
}  /* unm_host_async_accept */
